// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the Apache Public Source License 2.0 ("the License").
// You may not use this file except in compliance with the License.

package org.ibex.xt;
import org.ibex.js.*;
import org.ibex.util.*;
import org.ibex.io.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
//import com.thoughtworks.xstream.*;
//import org.prevayler.*;
//import org.prevayler.implementation.snapshot.*;

public class Servlet extends Prevalence.PrevalentServlet {

    public static final JS METHOD = new JS.Method();
    private ServletScope servletscope = null;
    private String path;

    public void doPost(HttpServletRequest request, HttpServletResponse response) throws IOException { doGet(request, response); }
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {
        servletscope = new ServletScope(request, response, cx);
        path = cx.getRealPath(((HttpServletRequest)request).getServletPath());
        Reader xmlreader = new InputStreamReader(new FileInputStream(path));
        Writer w = response.getWriter();
        new Template(servletscope, servletscope, xmlreader).wrap(null).toXML(w);
        w.flush();
    }

    // ServletScope //////////////////////////////////////////////////////////////////////////////

    public class ServletScope extends JSSubProperties {
        HttpServletRequest request;
        HttpServletResponse response;
        HttpSession session;
        ServletContext cx;
        public String getRealPath(String s) { return cx.getRealPath(s); }
        public ServletScope(ServletRequest request, ServletResponse response, ServletContext cx) {
            super();
            this.request = (HttpServletRequest)request;
            this.response = (HttpServletResponse)response;
            this.session = this.request.getSession();
            this.cx = cx;
        }

        // FIXME: setattributes, cookies
        //#repeat params/requestHeader/attributes \
        //        getParameter/getHeader/getAttribute \
        //        getParameterNames/getHeaderNames/getAttributeNames \
        //        request/request/session/response \
        //        setAttribute/setAttribute/setAttribute
        private JS params = new JS.Obj() {
                public JS get(JS key) throws JSExn { return JSU.S(request.getParameter(JSU.toString(key)).toString()); }
                public void put(JS key, JS val) throws JSExn { request.setAttribute(JSU.toString(key), JSU.toString(val)); }
                public Enumeration keys() throws JSExn {
                    return new JS.Enumeration.JavaStringEnumeration(request.getParameterNames()); }
            };
        //#end

        //case "prevalent":             return (JS)prevayler.prevalentSystem();
        public JS _get(JS key) throws JSExn {
            //#switch(JSU.toString(key))
            case "body":
            case "arg":                   return null;
            case "request":               return SUBPROPERTY;
            case "request.user":          return JSU.S(request.getRemoteUser());
            case "request.header":        return requestHeader;
            case "request.method":        return JSU.S(request.getMethod());
            case "request.remote":        return SUBPROPERTY;
            case "request.remote.ip":     return JSU.S(request.getRemoteAddr());
            case "request.remote.host":   return JSU.S(request.getRemoteHost());
            case "request.ssl":           return JSU.B(request.isSecure());
            case "request.path":          return JSU.S(request.getPathInfo());
            case "response":              return SUBPROPERTY;
            case "form":                  return SUBPROPERTY;
            case "form.fields":           return currentForm.fields();
            case "session":               return SUBPROPERTY;
            case "session.created":       return new JSDate(request.getSession(true).getCreationTime());
            case "session.accessed":      return new JSDate(request.getSession(true).getLastAccessedTime());
            case "session.invalidate":    return METHOD;
            case "page":                  return SUBPROPERTY;
            case "page.lastmodified":     return new JSDate(new File(path).lastModified());
            case "context":               return SUBPROPERTY;
            case "context.list":          return METHOD;
            case "params":                return params;
            //#end
                // cookie, responseheader needed, session.attributes
            return null;
        }

        Form currentForm = null;

        public void put(JS key, JS val) throws JSExn {
            try {
            //#switch(JSU.toString(key))
            case "response.code":         response.setStatus(JSU.toInt(val));
            case "response.redirect":     response.sendRedirect(JSU.toString(val));
            case "response.contentType":  response.setContentType(JSU.toString(val));
            case "transaction":
                /*
                try { prevayler.execute(new Prevalence.JSTransaction((JS)val));
                } catch (Exception e) { e.printStackTrace(); throw new RuntimeException(e); }
                */
            //#end
            } catch (IOException e) { throw new JSExn(JSU.S(e.toString())); }
        }

        public JS call(JS method, JS[] args) throws JSExn {
            //#switch(JSU.toString(method))
            case "session.invalidate":    request.getSession(true).invalidate(); return null;
            case "context.list":
                String path = JSU.toString(args[0]);
                if (path.indexOf("..") != -1) throw new JSExn("cannot use .. in paths");
                File f = new File(cx.getRealPath("/") + File.separatorChar + path);
                if (!f.isDirectory()) return null;
                String[] contents = f.list();
                JSArray ret = new JSArray(contents.length);
                for(int i=0; i<contents.length; i++) ret.put(JSU.N(ret.size()), JSU.S(contents[i]));
                return ret;
            //#end
            return null;
        }
    }

}
